# -*- coding:utf-8 -*-
from .utils import load


class Prop(object):

    def __init__(self,
                 _obj_name,
                 _name,
                 args,
                 ret_val=None,
                 ret_factory=None,
                 **kwargs):
        """

        :param _obj_name: 被mock的属性拥有者
        :param _name: 被mock的属性名称
        :param ret_val: mock的返回值
        :param ret_factory: mock的返回值生产工厂
        :param asyncable: 被mock的是属性是否是异步的
        :param callable: 被mock的属性是否是可调用的
        :param kwargs: mark传入的kwargs参数
        """
        self.obj = load(_obj_name)
        self.name = _name
        self.ret_val = ret_val
        self.ret_factory = load(ret_factory) \
            if isinstance(ret_factory, str) else ret_factory
        # 优先使用手动指定的标志位
        if "asyncable" in kwargs:
            self.asyncable = kwargs.pop("asyncable")
        if "callable" in kwargs:
            self.callable = kwargs.pop("callable")
        self.args = args or tuple()
        self.kwargs = kwargs or dict()

    def __iter__(self):
        yield self.obj
        yield self.name
        if self.callable:
            yield self
        else:
            if self.ret_factory:
                self.ret_val = self.ret_factory()
            yield self.ret_val

    def __call__(self, *args, **kwargs):
        kwargs.update(self.kwargs)
        if self.ret_factory:
            ret_val = self.ret_factory(*(args + self.args), **kwargs)
        else:
            ret_val = self.ret_val

        if self.asyncable:
            from .compact import get_coroutine
            return get_coroutine(ret_val)
        else:
            return ret_val

    def __str__(self):
        return "<Prop(name={name} obj={obj} asyncable={asyncable} " \
               "callable={callable} kwargs={kwargs} ret_val={ret_val} " \
               "ret_factory={ret_factory})>".format(**self.__dict__)

    __repr__ = __str__


def parse(mock, args, kwargs=None):
    """
    将mark转换成prop
    :param mock:
    :param args:
    :param kwargs:
    :return:
    """
    obj_name, _, prop = mock.rpartition(".")
    return Prop(obj_name, prop, args, **kwargs)



######为了发现在当前目录下创建的file包手动将当前目录加入环境变量######
import os
import sys
sys.path.insert(0, os.path.dirname(__file__))
######正常测试我们的项目应该已存在于pythonpath中所以不需要这么写#####
import pytest

from file.file import File


# 如果mock掉的方法不需要逻辑判断，直接使用ret_val指定返回值即可
@pytest.mark.usefixtures("mock")
@pytest.mark.prop("file.file.File.load", ret_val=File("直接mock返回值"))
def test_load1():
    assert File.load().filename == "直接mock返回值"


# 无论被mock的方法是异步还是同步，无脑指定返回值
# 不过异步的方法测试需要pytest-asyncio这个包支持，指定一个asyncio的mark才可以
@pytest.mark.asyncio
@pytest.mark.usefixtures("mock")
@pytest.mark.prop("file.file.File.load_async", ret_val=File("异步方法也能mock"))
async def test_load2():
    assert (await File.load_async()).filename == "异步方法也能mock"


# 如果我们用来替代的mock方法有一定的逻辑，我们可以指定一个ret_factory，指向替代方法的包地址
@pytest.mark.asyncio
@pytest.mark.usefixtures("mock")
@pytest.mark.prop("file.file.File.load_async", ret_factory="factories.mock_load")
async def test_load3():
    assert (await File.load_async()).filename == "通过工厂mock"


# 这种就是有业务逻辑的替代方法，他可以通过filename不同返回不同的File对象
@pytest.mark.usefixtures("mock")
@pytest.mark.prop("file.file.File.load", ret_factory="factories.mock_load", filename="还能给工厂传参")
def test_load4():
    assert File.load().filename == "还能给工厂传参"


# 装饰器太长怎么办，取个别名来缩短导包参数
from pytest_apistellar import prop_alias

file = prop_alias("file.file.File", "factories")


@pytest.mark.usefixtures("mock")
@file("load", ret_factory="mock_load", filename="使用别名装饰器，把前缀连起来")
def test_load4():
    assert File.load().filename == "使用别名装饰器，把前缀连起来"


# module作用域的mock
pytestmark = [
        file("load", ret_factory="mock_load", filename="这是一个module作用域的mock")
    ]


@pytest.mark.usefixtures("mock")
def test_load5():
    assert File.load().filename == "这是一个module作用域的mock"


@pytest.mark.usefixtures("mock")
class TestLoad(object):
    # class作用域的mock
    pytestmark = [
        file("load", ret_factory="mock_load", filename="这是一个class作用域的mock")
    ]

    def test_load6(self):
        assert File.load().filename == "这是一个class作用域的mock"

    # funtion作用域的mock
    @file("load", ret_factory="mock_load", filename="这是一个function作用域的mock")
    def test_load7(self):
        assert File.load().filename == "这是一个function作用域的mock"





class File(object):

    def __init__(self, filename):
        self.filename = filename

    @classmethod
    def load(cls):
        return File(filename="这个文件是从数据库获取的")

    @classmethod
    async def load_async(cls):
        return File(filename="这个文件是从数据库获取的")



# -*- coding:utf-8 -*-
from _pytest.mark import MarkDecorator, Mark


__version__ = "0.1.1"


class DecoratorProxy(object):
    """
    代理MarkDecorator，重载了其__call__方法，对prop_name和ret_factory进行前缀拼接操作。
    """
    def __init__(self, mock_obj_prefix, mock_factory_prefix, decorator):
        self.decorator = decorator
        self.mock_obj_prefix = mock_obj_prefix
        self.mock_factory_prefix = mock_factory_prefix

    def __call__(self, prop_name, *args, **kwargs):
        ret_factory = kwargs.pop("ret_factory", None)

        if self.mock_obj_prefix:
            pn = "{}.{}".format(self.mock_obj_prefix, prop_name)
        else:
            pn = prop_name

        if ret_factory and isinstance(ret_factory, str) and self.mock_factory_prefix:
            fn = "{}.{}".format(self.mock_factory_prefix, ret_factory)
        else:
            fn = ret_factory
        return self.decorator(pn, *args, ret_factory=fn, **kwargs)

    def __getattr__(self, item):
        return getattr(self.decorator, item)


def prop_alias(mock_obj_prefix=None, mock_factory_prefix="factories"):
    """
    prop mark装饰器别名，指定要mock对象及工厂对象名字的前缀，减少装饰器长度。
    :param mock_obj_prefix:
    :param mock_factory_prefix:
    :return:
    """
    return DecoratorProxy(mock_obj_prefix,
                          mock_factory_prefix,
                          MarkDecorator(Mark("prop", (), {})))



# -*- coding:utf-8 -*-
import re
import os
import six
import pytest
import inspect

from abc import ABCMeta, abstractmethod

from _pytest.mark import Mark
from _pytest.monkeypatch import MonkeyPatch

from .utils import load
from .parser import parse


class MarkerWrapper(object):
    """
    Mark类__eq__使用(name, args, kwargs)是否相同来判断，
    无法满足去重的要求，所以使用这个类来包装一下使用id来去重。
    """
    def __init__(self, marker):
        self.marker = marker

    def __eq__(self, other):
        return id(self.marker) == id(other)


@six.add_metaclass(ABCMeta)
class Patcher(object):

    @property
    @abstractmethod
    def name(self):
        """
        定义一个名字
        :return:
        """

    @property
    @abstractmethod
    def total_markers(self):
        """
        由于iter_markers会返回所有scope下的mark，
        所以使用这个属性来保证每个markers只处理一次，
        type: list
        :return:
        """

    def __init__(self, markers):
        self.monkey_patch = MonkeyPatch()
        self.markers = markers

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.monkey_patch.undo()

    def process(self):
        for mark in self.markers:
            if MarkerWrapper(mark) not in self.total_markers:
                self.process_mark(mark)
                self.total_markers.append(mark)

    @abstractmethod
    def process_mark(self, mark):
        """
        个性化定制mark处理方法
        :return:
        """

    @classmethod
    @abstractmethod
    def config_parse(cls, mark):
        """
        config解析方式
        :param mark:
        :return: Mark
        """

    @classmethod
    def from_pytestconfig(cls, pytestconfig):
        mocks = pytestconfig.inicfg.get(cls.name)
        ms = list()
        if mocks:
            for m in mocks.strip().split("\n"):
                if m.strip():
                    mark = cls.config_parse(m.strip())
                    if mark:
                        ms.append(mark)
        return cls(ms)

    @classmethod
    def from_request(cls, request, *args, **kwargs):
        return cls(request.node.iter_markers(cls.name))


class PropPatcher(Patcher):
    """
    用来monkey patch 属性
    """
    name = "prop"
    total_markers = list()

    def guess_attr(self, prop, old, mock, func):
        # 证明old是prop
        if hasattr(old, prop):
            setattr(mock, prop, getattr(old, prop))
        # 如果手动指定了，则使用手动指定的
        elif hasattr(mock, prop):
            pass
        # 否则，判定
        else:
            setattr(mock, prop, func(old))

    def process_mark(self, mark):
        mock = parse(mark.args[0], mark.args[1:], kwargs=mark.kwargs)
        try:
            old = getattr(mock.obj, mock.name, None)
            try:
                from asyncio import iscoroutinefunction
                self.guess_attr("asyncable", old, mock, iscoroutinefunction)
            except ImportError:
                mock.asyncable = False
            self.guess_attr("callable", old, mock, callable)
            # apistellar的依赖注入需要return 的signature
            if mock.callable:
                if getattr(old, "__annotations__", None):
                    if old.__annotations__["return"]:
                        mock.__signature__ = inspect.Signature(
                            return_annotation=old.__annotations__["return"])
                        mock.__annotations__ = {
                            "return": old.__annotations__["return"]}
        except AttributeError:
            raise RuntimeError("{} has not attr: {}".format(mock.obj, mock.name))
        self.monkey_patch.setattr(*mock, raising=False)

    @classmethod
    def config_parse(cls, mark):
        if "->" in mark:
            mark, ret_factory = mark.split("->", 1)
            kwargs = {"ret_factory": ret_factory}
        else:
            mark, ret_val = mark.split("=", 1)
            try:
                ret_val = eval(ret_val)
            except NameError:
                # 如果报错了，可能是字符串描述的模块没有导入，则导入模块
                index = ret_val.index("(")
                #如果存在(，则证明需要调用函数或类
                if index != -1:
                    prop_str, args_str = ret_val[:index], ret_val[index:]
                    prop = load(prop_str)
                    locals()[prop.__name__] = prop
                    ret_val = eval(prop.__name__ + args_str)
                else:
                    ret_val = load(ret_val)
            kwargs = {"ret_val": ret_val}

        return Mark(cls.name, tuple([mark]), kwargs)


class ItemPatcher(Patcher):
    """
    用来monkey patch Item
    """
    name = "item"
    total_markers = list()
    mark_config_regex = re.compile(r"(.+?)\[(.+?)\]\s*=\s*?(.+)")

    def process_mark(self, mark):
        for key, val in mark.kwargs.items():
            item = mark.args[0]
            if isinstance(item, str):
                item = load(item)
            self.monkey_patch.setitem(item, key, val)

    @classmethod
    def config_parse(cls, mark_str):
        mth = cls.mark_config_regex.search(mark_str)
        if mth:
            prop_name, key, val = mth.groups()
            return Mark(cls.name, tuple([prop_name]), {eval(key): eval(val)})


class EnvPatcher(Patcher):
    """
    用来monkey patch 环境变量
    """
    name = "env"
    total_markers = list()

    def process_mark(self, mark):
        prepend = mark.kwargs.pop("prepend", None)
        for key, val in mark.kwargs.items():
            self.monkey_patch.setenv(key, val, prepend)

    @classmethod
    def config_parse(cls, mark_str):
        key, val = mark_str.split("=", 1)
        return Mark(cls.name, tuple(), {key: val})


class PathPatcher(Patcher):
    """
        用来monkey patch 目录
        """
    name = "path"
    total_markers = list()

    def process_mark(self, mark):
        self.monkey_patch.chdir(mark.args[0])

    @classmethod
    def config_parse(cls, mark_str):
        return Mark(cls.name, tuple([mark_str]), dict())


class SysPathPatcher(PathPatcher):
    """
        用来monkey patch sys path
        """
    name = "syspath"
    total_markers = list()

    def process_mark(self, mark):
        self.monkey_patch.syspath_prepend(os.path.abspath(mark.args[0]))


def process(request, load_from="request"):
    with getattr(PathPatcher, "from_%s" % load_from)(request) as path_patcher,\
         getattr(SysPathPatcher, "from_%s" % load_from)(request) as syspath_patcher,\
         getattr(EnvPatcher, "from_%s" % load_from)(request) as env_patcher,\
         getattr(ItemPatcher, "from_%s" % load_from)(request) as item_patcher,\
         getattr(PropPatcher, "from_%s" % load_from)(request) as prop_petcher:
        path_patcher.process()
        syspath_patcher.process()
        env_patcher.process()
        item_patcher.process()
        prop_petcher.process()
        yield


def build(scope, load_from="request"):
    def mock(request, pytestconfig):
        gen = process(locals()[load_from], load_from=load_from)
        yield next(gen)

    return pytest.fixture(scope=scope, name="%s_mock" % scope)(mock)


# -*- coding:utf-8 -*-
import time
import socket
import warnings
import threading


def run_server(path, container, port=None):
    """
    创建一个简单的server用来测试
    :param path:
    :param port:
    :return:
    """
    try:
        import asyncio
        from apistellar import Application
        from uvicorn.main import Server, HttpToolsProtocol
    except ImportError:
        warnings.warn("Python3.6: apistellar required. ")
        raise
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    app = Application("test", current_dir=path)
    port = port or free_port()
    server = Server(app, "127.0.0.1", port, loop, None, HttpToolsProtocol)
    loop.run_until_complete(server.create_server())
    if server.server is not None:
        container.append(loop)
        container.append(server)
        loop.create_task(server.tick())
        loop.run_forever()


def create_server(path):
    container = []
    th = threading.Thread(target=run_server, args=(path, container))
    th.setDaemon(True)
    th.start()
    # 等待10秒
    for i in range(100):
        time.sleep(0.1)
        if len(container) == 2:
            break
    else:
        raise RuntimeError("子线程启动超时！")
    return container


def free_port():
    """
    Determines a free port using sockets.
    """
    free_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    free_socket.bind(('0.0.0.0', 0))
    free_socket.listen(5)
    port = free_socket.getsockname()[1]
    free_socket.close()
    return port


def load(prop_str):
    """
    返回字符串表示的模块、函数、类、若类的属性等
    :param prop_str: module1.class.function....
    :return: function
    """
    attr_list = []
    # 每次循环将prop_str当模块路径查找，成功则返回，
    # 失败则将模块路径回退一级，将回退的部分转换成属性
    # 至到加载模块成功后依次从模块中提取属性。
    ex = None

    while prop_str:
        try:
            obj = __import__(prop_str, fromlist=prop_str.split(".")[-1])
            for attr in attr_list:
                obj = getattr(obj, attr)
            return obj
        except (AttributeError, ImportError) as e:
            prop_str, _sep, attr_str = prop_str.rpartition('.')
            attr_list.insert(0, attr_str)
            ex = e
    else:
        raise ex
